@@ -1,6 +1,7 @@
 import json
 import logging
 import numbers
+from collections import defaultdict, namedtuple
 from datetime import date, datetime, time, timedelta
 from decimal import Decimal
 from warnings import warn
@@ -613,18 +614,24 @@ class ForeignKeyWidget(Widget):
         val = super().clean(value)
         if val:
             if self.use_natural_foreign_keys:
-                # natural keys will always be a tuple, which ends up as a json list.
-                value = json.loads(value)
-                return self.model.objects.get_by_natural_key(*value)
+                return self.get_instance_by_natural_key(value)
             else:
-                lookup_kwargs = self.get_lookup_kwargs(value, row, **kwargs)
-                obj = self.get_queryset(value, row, **kwargs).get(**lookup_kwargs)
+                obj = self.get_instance_by_lookup_fields(value, row, **kwargs)
                 if self.key_is_id:
                     return obj.pk
                 return obj
         else:
             return None
 
+    def get_instance_by_natural_key(self, value):
+        # natural keys will always be a tuple, which ends up as a json list.
+        value = json.loads(value)
+        return self.model.objects.get_by_natural_key(*value)
+
+    def get_instance_by_lookup_fields(self, value, row, **kwargs):
+        lookup_kwargs = self.get_lookup_kwargs(value, row, **kwargs)
+        return self.get_queryset(value, row, **kwargs).get(**lookup_kwargs)
+
     def get_lookup_kwargs(self, value, row, **kwargs):
         """
         :return: the key value pairs used to identify a model instance.
@@ -669,6 +676,138 @@ class ForeignKeyWidget(Widget):
         return value
 
 
+class _CachedQuerySetWrapper:
+    """
+    A wrapper around a Django QuerySet that caches its results in a dictionary
+    for quick lookups.
+
+    This class has the same 'get()' method signature as a QuerySet
+    because it is intended to be used as a drop-in replacement for QuerySet
+    in case of ForeignKeyWidget that calls 'get()' method for every row
+    in the import dataset.
+    """
+
+    def __init__(self, queryset):
+        self.queryset = queryset
+        self.model = queryset.model
+
+    def _get_instances(self, queryset, key_cls):
+        """
+        Converts a queryset into a dictionary mapping lookup fields values
+        to lists of instances. The values of the returned dict are lists
+        to handle potential multiple instances with the same lookup fields values.
+        """
+        if hasattr(self, "_instances"):
+            return self._instances
+
+        self._instances = defaultdict(list)
+        for instance in queryset:
+            key = key_cls(
+                **{field: getattr(instance, field) for field in key_cls._fields}
+            )
+            self._instances[key].append(instance)
+
+        return self._instances
+
+    def get(self, **lookup_fields):
+        Key = namedtuple("Key", list(lookup_fields.keys()))
+
+        instances = self._get_instances(self.queryset, Key)
+        key = Key(**lookup_fields)
+        result = instances.get(key, [])
+
+        if len(result) == 1:
+            return result[0]
+
+        if len(result) == 0:
+            raise self.model.DoesNotExist(
+                "%s matching query does not exist." % self.model._meta.object_name
+            )
+        raise self.model.MultipleObjectsReturned(
+            "get() returned more than one %s -- it returned %s!"
+            % (self.model._meta.object_name, len(result))
+        )
+
+
+class CachedForeignKeyWidget(ForeignKeyWidget):
+    """
+    A :class:`~import_export.widgets.ForeignKeyWidget` subclass that caches
+    the queryset results to minimize database hits during import. The default
+    :class:`~import_export.widgets.ForeignKeyWidget` makes query for each row,
+    which can be inefficient for large imports. This widget fetches all related
+    instances once and caches them in memory for subsequent lookups.
+
+    Using this class has some limitations:
+
+    - It does not support caching when ``use_natural_foreign_keys=True`` is set.
+
+    - It calls :meth:`~import_export.widgets.ForeignKeyWidget.get_queryset` only once,
+      so if the queryset depends on the row data, this widget may not work as expected.
+      You must be sure that the queryset is static for all rows.
+      Avoid using :class:`~import_export.widgets.CachedForeignKeyWidget`
+      in the following way::
+
+            class FullNameForeignKeyWidget(CachedForeignKeyWidget):
+                def get_queryset(self, value, row, *args, **kwargs):
+                    return self.model.objects.filter(
+                        first_name__iexact=row["first_name"],
+                        last_name__iexact=row["last_name"]
+                    )
+
+      It makes more sense to filter by static values::
+
+            class ActiveForeignKeyWidget(CachedForeignKeyWidget):
+                def get_queryset(self, value, row, *args, **kwargs):
+                    return self.model.objects.filter(active=True)
+
+    - It stores data in a hash table where the key is a tuple of the fields that
+      returned by :meth:`~import_export.widgets.ForeignKeyWidget.get_lookup_kwargs`.
+      You must be sure that the lookup fields are the same for all rows.
+      If the lookup fields differ between rows, this widget may not work as expected.
+      The following example is incorrect usage::
+
+            class MultiColumnForeignKeyWidget(CachedForeignKeyWidget):
+                def get_lookup_kwargs(self, value, row, **kwargs):
+                    if row['active'] == 'yes':
+                        return {self.field: value, 'active': True}
+                    else:
+                        return {self.field: value, 'inactive': True}
+
+    - It performs lookup on Python side, so the filtering logic
+      with non-text data types may not work::
+
+            class MultiColumnForeignKeyWidget(CachedForeignKeyWidget):
+                def get_lookup_kwargs(self, value, row, **kwargs):
+                    # row['birthday'] is a string like '01-01-2000'.
+                    #
+                    # It won't match the instance because the birthday values
+                    # in the cached instances are datetime objects, not strings.
+                    return {self.field: value, 'birthday': row['birthday']}
+
+    - It does not support complex lookups like ``__gt``, ``__lt``,
+      or filtering over relationships in the ``get_lookup_kwargs()``.
+      For example, the following code won't work::
+
+            class BookForeignKeyWidget(CachedForeignKeyWidget):
+                def get_lookup_kwargs(self, value, row, **kwargs):
+                    return {f'{self.field}__author': value}
+
+    :param model: The Model the ForeignKey refers to (required).
+    :param field: A field on the related model used for looking up a particular
+        object.
+    :param use_natural_foreign_keys: Use natural key functions to identify
+        related object, default to False
+    """
+
+    def get_instance_by_lookup_fields(self, value, row, **kwargs):
+        if not hasattr(self, "_cached_qs"):
+            queryset = self.get_queryset(value, row, **kwargs)
+            self._cached_qs = _CachedQuerySetWrapper(queryset)
+
+        lookup_kwargs = self.get_lookup_kwargs(value, row, **kwargs)
+        return self._cached_qs.get(**lookup_kwargs)
+
+
 class ManyToManyWidget(Widget):
     """
     Widget that converts between representations of a ManyToMany relationships
