@@ -5,13 +5,15 @@ from unittest import mock
 from unittest.mock import patch
 
 import django
+import tablib
 from core.models import Author, Book, Category
 from core.tests.utils import ignore_utcnow_deprecation_warning
+from django.db import connection
 from django.test import TestCase
-from django.test.utils import override_settings
+from django.test.utils import CaptureQueriesContext, override_settings
 from django.utils import timezone
 
-from import_export import widgets
+from import_export import fields, resources, widgets
 from import_export.exceptions import WidgetError
 
 
@@ -26,7 +28,7 @@ class WidgetTest(TestCase):
         self.assertEqual("1", self.widget.render(1))
 
 
-class RowDeprecationTestMixin(object):
+class RowDeprecationTestMixin:
     def test_render_row_deprecation(self):
         with self.assertWarnsRegex(
             DeprecationWarning,
@@ -109,6 +111,12 @@ class FormatDatetimeTest(TestCase):
         )
 
 
+class CustomDate(date):
+    """test derived instance of date"""
+
+    pass
+
+
 class DateWidgetTest(TestCase, RowDeprecationTestMixin):
     def setUp(self):
         self.date = date(2012, 8, 13)
@@ -117,6 +125,10 @@ class DateWidgetTest(TestCase, RowDeprecationTestMixin):
     def test_render(self):
         self.assertEqual(self.widget.render(self.date), "13.08.2012")
 
+    def test_render_derived_date(self):
+        derived_date = CustomDate(2012, 8, 13)
+        self.assertEqual(self.widget.render(derived_date), "13.08.2012")
+
     def test_render_none(self):
         self.assertEqual(self.widget.render(None), "")
 
@@ -163,6 +175,12 @@ class DateWidgetTest(TestCase, RowDeprecationTestMixin):
         self.assertEqual(("%Y-%m-%d",), self.widget.formats)
 
 
+class CustomDateTime(datetime):
+    """test derived instance of datetime"""
+
+    pass
+
+
 class DateTimeWidgetTest(TestCase, RowDeprecationTestMixin):
     def setUp(self):
         self.datetime = datetime(2012, 8, 13, 18, 0, 0)
@@ -171,6 +189,10 @@ class DateTimeWidgetTest(TestCase, RowDeprecationTestMixin):
     def test_render(self):
         self.assertEqual(self.widget.render(self.datetime), "13.08.2012 18:00:00")
 
+    def test_render_derived_datetime(self):
+        derived_datetime = CustomDateTime(2012, 8, 13, 18, 0, 0)
+        self.assertEqual(self.widget.render(derived_datetime), "13.08.2012 18:00:00")
+
     def test_render_none(self):
         self.assertEqual(self.widget.render(None), "")
 
@@ -267,6 +289,12 @@ class DateTimeWidgetBefore1900Test(TestCase):
         self.assertEqual(self.datetime, self.widget.clean("13.08.1868"))
 
 
+class CustomTime(time):
+    """test derived instance of time"""
+
+    pass
+
+
 class TimeWidgetTest(TestCase, RowDeprecationTestMixin):
     def setUp(self):
         self.time = time(20, 15, 0)
@@ -275,6 +303,10 @@ class TimeWidgetTest(TestCase, RowDeprecationTestMixin):
     def test_render(self):
         self.assertEqual(self.widget.render(self.time), "20:15:00")
 
+    def test_render_derived_time(self):
+        derived_time = CustomTime(20, 15, 0)
+        self.assertEqual(self.widget.render(derived_time), "20:15:00")
+
     def test_render_none(self):
         self.assertEqual(self.widget.render(None), "")
 
@@ -390,6 +422,22 @@ class FloatWidgetTest(TestCase, RowDeprecationTestMixin):
     def test_clean(self):
         self.assertEqual(self.widget.clean(11.111), self.value)
 
+    @override_settings(USE_THOUSAND_SEPARATOR=True)
+    def test_clean_numeric_separators(self):
+        self.assertEqual(self.widget.clean("1,234.5"), 1234.5)
+
+    @override_settings(LANGUAGE_CODE="ar", USE_THOUSAND_SEPARATOR=True)
+    def test_clean_numeric_separators_arabic(self):
+        self.assertEqual(self.widget.clean("1.234,5"), 1234.5)
+
+    @override_settings(LANGUAGE_CODE="zh-hans", USE_THOUSAND_SEPARATOR=True)
+    def test_clean_numeric_separators_chinese_simplified(self):
+        self.assertEqual(self.widget.clean("1234.5"), 1234.5)
+
+    @override_settings(LANGUAGE_CODE="fr", USE_THOUSAND_SEPARATOR=True)
+    def test_clean_numeric_separators_french(self):
+        self.assertEqual(self.widget.clean("1\xa0234,5"), 1234.5)
+
     def test_render(self):
         self.assertEqual(self.widget.render(self.value), "11.111")
 
@@ -419,6 +467,22 @@ class DecimalWidgetTest(TestCase, RowDeprecationTestMixin):
         self.assertEqual(self.widget.clean("11.111"), self.value)
         self.assertEqual(self.widget.clean(11.111), self.value)
 
+    @override_settings(USE_THOUSAND_SEPARATOR=True)
+    def test_clean_numeric_separators(self):
+        self.assertEqual(self.widget.clean("1,234.5"), Decimal("1234.5"))
+
+    @override_settings(LANGUAGE_CODE="ar", USE_THOUSAND_SEPARATOR=True)
+    def test_clean_numeric_separators_arabic(self):
+        self.assertEqual(self.widget.clean("1.234,5"), Decimal("1234.5"))
+
+    @override_settings(LANGUAGE_CODE="zh-hans", USE_THOUSAND_SEPARATOR=True)
+    def test_clean_numeric_separators_chinese_simplified(self):
+        self.assertEqual(self.widget.clean("1234.5"), Decimal("1234.5"))
+
+    @override_settings(LANGUAGE_CODE="fr", USE_THOUSAND_SEPARATOR=True)
+    def test_clean_numeric_separators_french(self):
+        self.assertEqual(self.widget.clean("1\xa0234,5"), Decimal("1234.5"))
+
     def test_render_coerce_to_string_is_False(self):
         self.widget = widgets.DecimalWidget(coerce_to_string=False)
         self.assertEqual(self.widget.render(self.value), self.value)
@@ -465,6 +529,22 @@ class IntegerWidgetTest(TestCase, RowDeprecationTestMixin):
         self.assertEqual(self.widget.clean(" "), None)
         self.assertEqual(self.widget.clean("\n\t\r"), None)
 
+    @override_settings(USE_THOUSAND_SEPARATOR=True)
+    def test_clean_numeric_separators(self):
+        self.assertEqual(self.widget.clean("1,234.5"), 1234)
+
+    @override_settings(LANGUAGE_CODE="ar", USE_THOUSAND_SEPARATOR=True)
+    def test_clean_numeric_separators_arabic(self):
+        self.assertEqual(self.widget.clean("1.234,5"), 1234)
+
+    @override_settings(LANGUAGE_CODE="zh-hans", USE_THOUSAND_SEPARATOR=True)
+    def test_clean_numeric_separators_chinese_simplified(self):
+        self.assertEqual(self.widget.clean("1234.5"), 1234)
+
+    @override_settings(LANGUAGE_CODE="fr", USE_THOUSAND_SEPARATOR=True)
+    def test_clean_numeric_separators_french(self):
+        self.assertEqual(self.widget.clean("1\xa0234,5"), 1234)
+
     def test_render_invalid_type(self):
         self.assertEqual(self.widget.render("a"), "")
 
@@ -533,7 +613,7 @@ class ForeignKeyWidgetTest(TestCase, RowDeprecationTestMixin):
         self.assertEqual(target_author, res)
 
     def test_render_handles_value_error(self):
-        class TestObj(object):
+        class TestObj:
             @property
             def attr(self):
                 raise ValueError("some error")
@@ -594,6 +674,191 @@ class ForeignKeyWidgetTest(TestCase, RowDeprecationTestMixin):
         )
 
 
+class CachedForeignKeyWidgetTest(TestCase, RowDeprecationTestMixin):
+    def setUp(self):
+        self.widget = widgets.CachedForeignKeyWidget(Author)
+        self.natural_key_author_widget = widgets.CachedForeignKeyWidget(
+            Author, use_natural_foreign_keys=True
+        )
+        self.natural_key_book_widget = widgets.CachedForeignKeyWidget(
+            Book, use_natural_foreign_keys=True
+        )
+        self.author = Author.objects.create(name="Foo")
+        self.book = Book.objects.create(name="Bar", author=self.author)
+
+    def tearDown(self):
+        if hasattr(self.widget, "_cached_qs"):
+            del self.widget._cached_qs
+        if hasattr(self.natural_key_author_widget, "_cached_qs"):
+            del self.natural_key_author_widget._cached_qs
+        if hasattr(self.natural_key_book_widget, "_cached_qs"):
+            del self.natural_key_book_widget._cached_qs
+
+    def test_clean(self):
+        self.assertEqual(self.widget.clean(self.author.id), self.author)
+
+    def test_clean_empty(self):
+        self.assertEqual(self.widget.clean(""), None)
+
+    def test_render(self):
+        self.assertEqual(self.widget.render(self.author), self.author.pk)
+
+    def test_render_empty(self):
+        self.assertEqual(self.widget.render(None), "")
+
+    def test_cache_hit(self):
+        author2 = Author.objects.create(name="Baz")
+        with CaptureQueriesContext(connection) as ctx:
+            self.assertEqual(self.widget.clean(self.author.id), self.author)  # query
+            self.assertEqual(self.widget.clean(author2.id), author2)  # cache hit
+            self.assertEqual(len(ctx.captured_queries), 1)
+
+        with CaptureQueriesContext(connection) as ctx:
+            self.assertEqual(
+                self.widget.clean(self.author.id), self.author
+            )  # cache hit
+            self.assertEqual(self.widget.clean(author2.id), author2)  # cache hit
+            self.assertEqual(len(ctx.captured_queries), 0)
+
+    def test_cache_is_not_shared_for_different_resource_instances(self):
+        class BookResource(resources.ModelResource):
+            author = fields.Field(
+                attribute="author",
+                column_name="Author",
+                widget=widgets.CachedForeignKeyWidget(Author),
+            )
+
+            class Meta:
+                model = Book
+
+        headers = ["Author"]
+        row = [self.author.id]
+        dataset = tablib.Dataset(row, headers=headers)
+        resource = BookResource()
+
+        self.assertFalse(hasattr(resource.fields["author"].widget, "_cached_qs"))
+        resource.import_data(dataset, dry_run=True)
+        self.assertTrue(hasattr(resource.fields["author"].widget, "_cached_qs"))
+
+        resource2 = BookResource()  # new instance. The cache must be reset
+        self.assertFalse(hasattr(resource2.fields["author"].widget, "_cached_qs"))
+
+    def test_clean_multi_column(self):
+        class BirthdayWidget(widgets.CachedForeignKeyWidget):
+            def get_queryset(self, value, row, *args, **kwargs):
+                return self.model.objects.filter(birthday=row["birthday"])
+
+        author2 = Author.objects.create(name="Foo")
+        author2.birthday = "2016-01-01"
+        author2.save()
+        birthday_widget = BirthdayWidget(Author, "name")
+        row_dict = {"name": "Foo", "birthday": author2.birthday}
+        self.assertEqual(birthday_widget.clean("Foo", row=row_dict), author2)
+
+    def test_invalid_get_queryset(self):
+        class BirthdayWidget(widgets.CachedForeignKeyWidget):
+            def get_queryset(self, value, row):
+                return self.model.objects.filter(birthday=row["birthday"])
+
+        birthday_widget = BirthdayWidget(Author, "name")
+        row_dict = {"name": "Foo", "age": 38}
+        with self.assertRaises(TypeError):
+            birthday_widget.clean("Foo", row=row_dict, row_number=1)
+
+    def test_lookup_multiple_columns(self):
+        # issue 1516 - override the values used to lookup an entry
+        class BookWidget(widgets.CachedForeignKeyWidget):
+            def get_lookup_kwargs(self, value, row, *args, **kwargs):
+                return {"name": row["name"], "author_email": row["authoremail"]}
+
+        target_book = Book.objects.create(
+            name="Baz", author=self.author, author_email="baz@email.com"
+        )
+        row_dict = {"name": "Baz", "authoremail": "baz@email.com"}
+        book_widget = BookWidget(Book, "name")
+        # prove that the overridden kwargs identify a row
+        res = book_widget.clean("non-existent name", row=row_dict)
+        self.assertEqual(target_book, res)
+
+    def test_render_handles_value_error(self):
+        class TestObj:
+            @property
+            def attr(self):
+                raise ValueError("some error")
+
+        t = TestObj()
+        self.widget = widgets.CachedForeignKeyWidget(mock.Mock(), "attr")
+        self.assertIsNone(self.widget.render(t))
+
+    def test_multiple_objects_error(self):
+        Author.objects.create(name=self.author.name)  # an author with duplicated name
+        widget = widgets.CachedForeignKeyWidget(Author, "name")
+
+        with self.assertRaises(Author.MultipleObjectsReturned) as e:
+            widget.clean(self.author.name)
+
+        self.assertEqual(
+            str(e.exception), "get() returned more than one Author -- it returned 2!"
+        )
+
+    def test_object_does_not_exist_error(self):
+        with self.assertRaises(Author.DoesNotExist) as e:
+            self.widget.clean("non-existent-id")
+
+        self.assertEqual(str(e.exception), "Author matching query does not exist.")
+
+    def test_author_natural_key_clean(self):
+        """
+        Ensure that we can import an author by its natural key. Note that
+        this will always need to be an iterable.
+        Generally this will be rendered as a list.
+        """
+        self.assertEqual(
+            self.natural_key_author_widget.clean(json.dumps(self.author.natural_key())),
+            self.author,
+        )
+
+    def test_author_natural_key_render(self):
+        """
+        Ensure we can render an author by its natural key. Natural keys will always be
+        tuples.
+        """
+        self.assertEqual(
+            self.natural_key_author_widget.render(self.author),
+            json.dumps(self.author.natural_key()),
+        )
+
+    def test_book_natural_key_clean(self):
+        """
+        Use the book case to validate a composite natural key of book name and author
+        can be cleaned.
+        """
+        self.assertEqual(
+            self.natural_key_book_widget.clean(json.dumps(self.book.natural_key())),
+            self.book,
+        )
+
+    def test_book_natural_key_render(self):
+        """
+        Use the book case to validate a composite natural key of book name and author
+        can be rendered
+        """
+        self.assertEqual(
+            self.natural_key_book_widget.render(self.book),
+            json.dumps(self.book.natural_key()),
+        )
+
+    def test_natural_foreign_key_with_key_is_id(self):
+        with self.assertRaises(WidgetError) as e:
+            widgets.CachedForeignKeyWidget(
+                Author, use_natural_foreign_keys=True, key_is_id=True
+            )
+        self.assertEqual(
+            "use_natural_foreign_keys and key_is_id " "cannot both be True",
+            str(e.exception),
+        )
+
+
 class ManyToManyWidget(TestCase, RowDeprecationTestMixin):
     def setUp(self):
         self.widget = widgets.ManyToManyWidget(Category)
@@ -602,21 +867,21 @@ class ManyToManyWidget(TestCase, RowDeprecationTestMixin):
         self.cat2 = Category.objects.create(name="Cat 2")
 
     def test_clean(self):
-        value = "%s,%s" % (self.cat1.pk, self.cat2.pk)
+        value = f"{self.cat1.pk},{self.cat2.pk}"
         cleaned_data = self.widget.clean(value)
         self.assertEqual(len(cleaned_data), 2)
         self.assertIn(self.cat1, cleaned_data)
         self.assertIn(self.cat2, cleaned_data)
 
     def test_clean_field(self):
-        value = "%s,%s" % (self.cat1.name, self.cat2.name)
+        value = f"{self.cat1.name},{self.cat2.name}"
         cleaned_data = self.widget_name.clean(value)
         self.assertEqual(len(cleaned_data), 2)
         self.assertIn(self.cat1, cleaned_data)
         self.assertIn(self.cat2, cleaned_data)
 
     def test_clean_field_spaces(self):
-        value = "%s, %s" % (self.cat1.name, self.cat2.name)
+        value = f"{self.cat1.name}, {self.cat2.name}"
         cleaned_data = self.widget_name.clean(value)
         self.assertEqual(len(cleaned_data), 2)
         self.assertIn(self.cat1, cleaned_data)
@@ -648,11 +913,11 @@ class ManyToManyWidget(TestCase, RowDeprecationTestMixin):
     def test_render(self):
         self.assertEqual(
             self.widget.render(Category.objects.order_by("id")),
-            "%s,%s" % (self.cat1.pk, self.cat2.pk),
+            f"{self.cat1.pk},{self.cat2.pk}",
         )
         self.assertEqual(
             self.widget_name.render(Category.objects.order_by("id")),
-            "%s,%s" % (self.cat1.name, self.cat2.name),
+            f"{self.cat1.name},{self.cat2.name}",
         )
 
     def test_render_value_none_as_blank(self):
